% function [BRDF, rOut] = FCBRDF( Esrc, xSrc, L, rOut, dr, powerDist )
% BRDF of field Esrc(xSrc,xSrc) after propagating L, at r = dr:dr:rOut 
% if powerDist is given and true, BRDF is power density
%
% May 24, 2020
%   First release in GIT. This uses singleFFT
%
function [BRDF, rOut] = FCBRDF( Esrc, xSrc, L, rOut, dr, powerDist )
if ~exist( 'dr', 'var' )
    dr = 0.001;
end
N = floor( rOut/dr + 0.001 );
xList = (-N:N)*dr;
rOut = xList(ceil(end/2):end);
[Eout, xOut] = singleFFT( L, Esrc, xSrc, xSrc, xList, xList );
% cannot interpolate out of range
rOut( rOut > xOut(end) ) = [];
% Eout(x,y) to Pout(r)
Pout = UtilTK.xy2r( xOut(1,:), abs(Eout).^2, rOut );
Ptot = sum(sum(abs(Esrc).^2))*(xSrc(2)-xSrc(1))^2;
% conversion from energy densiy to BRDF
if exist( 'powerDist', 'var' ) && powerDist
    BRDF = Pout;
else
    BRDF = Pout * L^2 / Ptot;
end
