%
% propagator : connects two ports
%
% May 11, 2020
%   The change on March 31st is incorrect. Amplitudes need to be scale by the ratio of the sizes, not squared.
%
% March 31, 2020
%   function outField = propField( obj, inField, kSB, keepPropSize )
%   when propL0 is 0 and the size is doubled by keepPropSize, the ampitude had to be quadrapled
%
% October 2, 2019
%   propOnDemand has a bug. The internal field depends on phiCorr of the field.
%   To support different length, internal field needs to be corrected for each length.
% 
% March 27, 2018
%   By choosing phiCorr(in) = (ratio - 1)/Lprop, Amap can be elimiated
%
% March 8, 2018
%   function [inPort, outPort] = HR2HRProp( obj ) added
%   This returns HR port to HR port connected by this propagator
%
% February 9, 2018
%   function kLGouyPhase added which returns k L - gouyPhase
%
% January 31, 2018
% function phase = totalPhase( obj, propID, subtractLockedPhase ) added
%     calculate phase change by a propagator, -kL + (n+m+1)gouyPhase, optionally subtract locked phase
% printInfo uptaed to include totalPhase change
%
% September 5th, 2017
%    PropObj has nRef as an argument
%    printInfo includes propABCD
%
% April 15, 2017
%    propOnDemand, propField and setupMisc have been updated to handle getField(ID, Wout, Lprop) properly.
%
% March 28, 2017
%    setupMisc modified to fix the calculation of Lprop.
%    obj.propL0 = londitudinal optical distance
%               = distance between mirror + mirror thickness * nRef
%    Lprop (used in this function ) = transversal optical distance
%               = distance between mirror + mirror thickness / nRef
%
% March 24,2017
%    resetDelL use the lockPhase when no phase is given in the argument
%
% March 13, 2017
%    resetDelL has a new argument, phase, and delL is chosen to make the one path phase to be this given phase
%    -(k0 + kSB)(L0 + delL) + (n+m) gouy00 = phase
%
% March 12, 2017
%    propField : if propL0 == 0, propField returns after coping input to output
%
% March 6, 2017
%    setupMisc and propField are modified.
%    Nmul, a new property is added.
%
%    When setupMisc is called, if Nprop is not specified in the argument,
%    Pmap, the propagation matrix, is enlarged to NFFT * Nmul.
%    When propField is called and if Nprop is not specified in the argument,
%    input field is expanded to NFFT * Nmul (size of Pmap) by filling 0 outside of NFFTxNFFT,
%    and the field is shrunk after propagation to the original size.
%    At any time, setupMisc can be called to change the internal propagator size to enlarge
%    the null region to suppress the aliasing.
%    Pmap can be swapped to change the propagation size.
%
% March 2, 2017
%    printInfo port ID is prepended by name
%
% February 13, 2017
%    propObj constructor has lockPhase as input argument
%
% September 2nd, 2015
%    ifo is added as a property
%    several functions with optics and fields as input arguments are
%    modified to use obj.ifo.optics and .fields
%
% April 9, 2015
%    the sign of the ratio for the adaption gridding is chosen
%    so that the extra phase of Amap becomes small
%           Amap = exp( 1i * k/2 * ( (ratio-1)/Lprop - getInvRoC(inPortID) ) * r^2 )
%
% September 18, 2014
%   setupMisc( obj, optics, fields, Lprop, sinkWfft, Nprop )
%   outField = propField( obj, inField, kSB, Nprop )
%   [outField, xl, yl] = propOnDemand( obj, optics, fields, Wsink, Lprop )
%
%   These are updated to support getField( propagation distance and output window size )
%   There is some issue in propOnDemand. Some sink window size does not work well, check out WHYWHYWHY
%
% August 29, 2014
%	function [outField, xl, yl] = propOnDemand( obj, optics, fields, Wout, Lprop )
%   When Lprop > 0, the propagation distance is changed
%   For this case, if Wout == 0, the window size is scaled proportional to Lprop
%
% September 11, 2013  Hiro Yamamoto
%
%
classdef PropObj < handle
    
    properties (SetObservable)
        % macroscopic length, integer multiple of the base wavelength
        L0 = NaN;
        % refractive index
        nRef = 1;
    end
    
    properties
        % microscpic length. Total length = L0 + delL
        delL = 0;
        
        % the extra phase for SB is calculated by exp( i kSB propL0 )
        propL0 = NaN;
        
        % input and output field IDs
        inPortID = NaN;
        outPortID = NaN;
        
        % coordinate scaling
        ratio = 1;  % Wfft(sink) / Wfft(src)  x -1 when the curvature changes sign
        CLa = 0;    % (1 - ratio) / (ratio * L0);
        
        % map to be multipled on field before propagation (A)
        % Amap is 1 by choosing proper phiCorr(in)
%        Amap = zeros(0,0); 
        
        % FFT Propagation matrix (K)
        % pmap is swapped, 0~f/2, -f/2,0, so
        % outField = fftshift( ifft2( Pmap * fft2( ifftshift( inField ))
        Pmap = zeros(0,0);
        
        % name of prop
        name = 'unnamed';
        
        % propagator ABCD
        propABCD
        
        % gouy phase
        gouy00
        
        % lock phase
        lockPhase  % one trip phase for this cavity
        
        % prooagation uses (NfftxNmul) x (NfftxNmul) by expanding the input and shrining the putput
        Nmul
        
        % IFO object
        ifo
    end
    
    % function to replace a static variable
    methods
        % constructor
        function obj = PropObj( name, L0, delL, lockPhase, nRef, inPortID, outPortID, ifo )
            obj.ifo = ifo;
            
            obj.outPortID = outPortID;
            obj.inPortID = inPortID;
            obj.nRef = nRef;
            obj.delL = delL;
            obj.L0 = L0;    % this updates propABCD
            obj.name = name;
            obj.lockPhase = lockPhase;
            obj.Nmul = 1;
            
            addlistener(obj, 'nRef', 'PostSet', @obj.updateNdep);
        end
        
        %%
        function Wout = setupMisc( obj, Wout, Lprop )
            % calculate parameters used for field progation
            %
            %   adaptive grid size is used for beam size change, srcWfft and sinkWfft are used for the ratio
            %   If the waist is passed during this propagation, ratio = -ratio, sort of slipping up-down/right-left
            %
            %   propOnDemand calls setupMiac with Wout and Lprop. Actual Wout could be different and is returned
            %
            % parameters calculated by this function
            %   Amap, Pmap : out = ifft( Pmap * fft( Amap * Ein ) )
            %   ratio
            %   gouy00
            %   CLa = -(1-1/ratio) / Lprop
            %       this will be used to convert internal format Eout to standard format
            %   For the normal call, i.e., no Wout nor Lprop, phiCorr in field objects at each side of the propagation
            %   For the onDemand call, this phase part is handled manually
            %
            %  Propagation is done using internal format and the conversion from internal to stanfdard format is as follows
            %       Eout = ifft( Pmap * fft( Amap * Ein ) ) * exp( -i (k0 * dL + kSB * L0) )
            %       Enominal = Einternal * exp( -i k/2 r^2 phiCorr ) / Wfft(field)
            %
            %       size of Ein, Eout, Amap : Nfft/2 x Nfft/2 (Nfft is the standard Nfft)
            %       size of Pmap : Nprop x Nprop (Nprop can be larger than Nfft for larger beam propagation)
            %
            %       Wout and Lprop are given only for propOnDemand
            %       Lprop is the propagation distance, 0 is zero, no special meaning
            %
            %       If Wout > 0 is specified, the field is calculated in (Wout x Wout) area
            %          Wout is the visible window size, and to be doubled to suppress aliasing
            %       If Wout = 0, Wfft is set to be optimal based on the beam size
            %       If Wout = -1, Wfft is set to be the same as the window size of the sink port of the propagator.
            %
            %       1) If Lprop == obj.L0, Wout/obj.Wsink is used to calculate the scaling, Nratio, power of 2,
            %          and newWsrc = obj.Wsrc * Nratio
            %       2) If Lprop ~= obj.L0, beam size ratio at the src and sink are used to calculate
            %          newWsrc = Wout / beam size(sink) * beam size(src), and newWsrc / obj.Wsrc is used
            %          to calculate Nratio (power of 2), and the final newWsrc = Wsrc * Nratio.
            %
            
            % Nratio is defined for on demand propagation
            onDemand = exist('Wout', 'var');
            
            % information about the src and sink of this prop
            % Wfft are values before aliasMargin is multiplied ...
            srcOpt = obj.ifo.optics.findOptic( obj.inPortID );
            srcWfft = obj.ifo.optics.optList( srcOpt ).Wfft / FFTTK.aliasMargin;
            sinkOpt = obj.ifo.optics.findOptic(obj.outPortID);
            sinkWfft = obj.ifo.optics.optList( sinkOpt ).Wfft / FFTTK.aliasMargin;

            if onDemand
                % if Wout or Lprop is specified
                onSink = abs(Lprop / obj.L0 - 1) < 1e-6; % propagation is the same distance as L0
                if onSink && (Wout < srcWfft )
                    Wout = sinkWfft;
                end
                % calculate the beam sizes at the src and sink using invQ
                [z,z0] = MMTK.q2zz0( 1/obj.ifo.fields.fieldList(obj.inPortID).invQ );
                wAtSrc = MMTK.qinv2W( 1/( z + 1i*z0 ), obj.ifo.fields.kCR );
                wAtSink = MMTK.qinv2W( 1/( z + Lprop + 1i*z0 ), obj.ifo.fields.kCR );
                wRatio = wAtSink / wAtSrc;
                if Wout == 0
                    % optmal value, i.e., srcWfft * beam size at sink / beam size at src
                    % the adaptive scaling
                    Wout = srcWfft * wRatio;
                elseif Wout < 0
                    Wout = sinkWfft;
                end
                % Wout has been set
                % Wout / wRatio is the new srcWfft which matches with Wout
                Nratio = Wout / wRatio / srcWfft;
                % if the change of the srcWindow is large
                if Nratio > 1.01
                    % change should be power of 2
                    Nratio = 2^ceil( log(Nratio)/log(2) );
                    % this is the new propagator grids, including aliasMargin
                    Nprop = obj.ifo.optics.ifo.ffts.Nfft * Nratio;
                    % this is the new srcWindow using Nration, 2^n
                    srcWfft = srcWfft * Nratio;
                    % Wout which is srcW * wRatio
                    Wout = srcWfft * wRatio;
                else
                    Nprop = obj.ifo.optics.ifo.ffts.Nfft;
                end 
                
                obj.propL0 = Lprop;
            else
                srcWfft = srcWfft * obj.Nmul;
                sinkWfft = sinkWfft * obj.Nmul;
                Wout = sinkWfft;
                Nprop = obj.ifo.optics.ifo.ffts.Nfft * obj.Nmul;

                % propL0 is used for calculate longitudinal phase for SB by propL0 * kRF, L(mirror-mirror) + thickness * n
                % Lprop is used in this function to calculate the transversal effct, L(mirror-mirror) + thicness / n
                obj.propL0 = obj.L0 + obj.ifo.optics.getDeapth( obj.inPortID, 1 ) + obj.ifo.optics.getDeapth( obj.outPortID, 1 );
                Lprop = obj.L0 + obj.ifo.optics.getDeapth( obj.inPortID, -1 ) + obj.ifo.optics.getDeapth( obj.outPortID, -1 );
            end
            
            obj.ratio = Wout / srcWfft;
            
            % sign of ratio is arbitrary, but ...
            % out = ifft( Pmap * fft( Amap * Ein ) )
            % Amap = exp( 1i * k/2 * ( (ratio-1)/Lprop - fields.getInvRoC(obj.inPortID) ) * r^2 )
            % so if ratio = 1 + L/R, phase correction is minial
            % Choose the sign of ratio to be the sign of (1 + Lprop*fields.getInvRoC(obj.inPortID))
            if (1 + Lprop*obj.ifo.fields.getInvRoC(obj.inPortID)) < 0
                obj.ratio = -obj.ratio;
            end
            
            if Lprop == 0
                % if no propagation, just identify matrix, with proper size ...
                obj.gouy00 = 0;
                
                % this case should not happen or used
                obj.CLa = 0;
                
                % when there is no propagation, just fill the RoC
                if ~onDemand
                    obj.ifo.fields.fieldList(obj.inPortID).phiCorr = obj.ifo.fields.getInvRoC(obj.inPortID);
                    obj.ifo.fields.fieldList(obj.outPortID).phiCorr = obj.ifo.fields.getInvRoC(obj.outPortID);
                end
                
                % these will never be used
                obj.Pmap = Nprop;
%                obj.Amap = 1;
            else
%                obj.gouy00 = MMTK.gouyL( 1/obj.ifo.fields.fieldList(obj.inPortID).invQ, Lprop / obj.nRef );
                obj.gouy00 = MMTK.gouyL( 1/obj.ifo.fields.fieldList(obj.inPortID).invQ, Lprop );
                
                % calculate maps for the FFT propagation
                % outField(x,y) = ifft( Pmap * fft( Amap * inField ) )
                %   inField internal  = Enominal * exp( i k/2 r^2 invRoC ) Wfft(in)
                %   outField internal = Enominal * exp( -i k/2 r^2 CLa ) Wfft(out)
                
                obj.CLa = -(1 - 1/obj.ratio) / Lprop;
                
                % Enominal = Einternal * exp( -i k/2 r^2 phiCorr ) / Wfft(field)
                
                % The FOG defines CL (=invRoc) for the going out field to be opposite to going in field
                % That makes this definition opposite, but this and FOG are consistent
                if ~onDemand
                    % for onDemand case, these factors are handled by hand, and does not need to store in field
%                    obj.ifo.fields.fieldList(obj.inPortID).phiCorr = obj.ifo.fields.getInvRoC(obj.inPortID); % 3/27/2018
                    obj.ifo.fields.fieldList(obj.inPortID).phiCorr = -obj.CLa * obj.ratio; % (ration - 1)/L
                    obj.ifo.fields.fieldList(obj.outPortID).phiCorr = -obj.CLa;
                end
                
                % Amap is the same size as original, so it can be multiplied with Ein
                % Pmap is Nnew x Nnew, using larger source window, filled with 0 out of the original source window
                % out = ifft( Pmap * fft( Amap * Ein ) )
                 % 3/27/2018   by choosing proper phiCorr(in), Amap = 1
                % obj.Amap = 1;
%                 obj.Amap = exp( 1i * obj.ifo.fields.kCR/2 * ...
%                     ( (obj.ratio-1)/Lprop - obj.ifo.fields.getInvRoC(obj.inPortID) ) * ...
%                     obj.ifo.optics.getRsq( srcOpt ) );
                
                Fvec = 1/(2* srcWfft*FFTTK.aliasMargin)*(-Nprop:2:(Nprop-2));
                [fx,fy] = meshgrid( Fvec, Fvec );
                obj.Pmap = exp( 1i * pi * obj.ifo.fields.lamCR * Lprop / obj.ratio * (fx.^2 + fy.^2) );
                Ncut = obj.FFTFiltC( srcWfft*FFTTK.aliasMargin, Wout*FFTTK.aliasMargin, ...
                    Lprop, Nprop, obj.ifo.fields.lamCR );
                obj.Pmap(Ncut) = 0;
                obj.Pmap = ifftshift(obj.Pmap);
            end
        end
        
        %%
        function outField = propField( obj, inField, kSB, keepPropSize )
            %
            % FFT propagation with magnification, was R_diffrC3 in FOG
            %	Eout = ifft( Pmap * fft( Amap * Ein ) ) * exp( -i (k0 * dL + kSB * L0) )
            %
            % inField and outField are in internal format
            % field size changes from the size of inField (Nfft) to NProp, the size of Pmap.
            % If keepNPropSize is false, the outField size after propagation is reduced to that of inField.
            % If keepNPropSize is true, the outfield size is the one determined by Pmap.
            %
            
            % if no propagation, just copy the input to output
            keepPropSize = exist('keepPropSize', 'var' ) && keepPropSize;
            if obj.propL0 == 0
                if keepPropSize
                    % * 4 is to account for the change of size by 2 x 2
                    outField = UtilTK.expand2x2( inField, obj.Pmap/FFTTK.aliasMargin ) * obj.Pmap/FFTTK.aliasMargin/size(inField,2) ;
                else
                    outField = inField;
                end
                return;
            end
            
            %   Nprop x Nprop is used for the FFT propagation in frequemncy space to calculate larger field props
            %   Default is twice of the field size
            %
            % scaleN is the difference of the field normalization due to the change of Nfft of inField and outField
            Nprop = size( obj.Pmap, 2 );
            if keepPropSize
                scaleN = Nprop / (FFTTK.aliasMargin * size( inField, 2 ));
                outDim = Nprop / FFTTK.aliasMargin;
            else
                scaleN = 1;
                outDim = size( inField, 2 );
            end
            
            %        expand input field to the size of propagator, Pmap, after multiplying correction factor, Amap
            %             myfield = UtilTK.expand2x2( obj.Amap .* inField, Nprop );
            %        convert to frequency domain
            %             Field_FFT = fft2( ifftshift(myfield) );
            %        propagate by Pmap, which has been swapped (0~f,-f~0)
            %             Field_FFT = obj.Pmap .* Field_FFT;
            %        move back to x-y domain
            %             outField = fftshift( ifft2( Field_FFT ) );
            %        remove outside zone used through propagation
            %             outField = UtilTK.useQuarter( outField, outDim );
            %        all in one line, to save a little bit of time
            outField = UtilTK.useQuarter( ...
                fftshift( ifft2( obj.Pmap .* ...
                fft2( ifftshift( ...
                UtilTK.expand2x2( inField, Nprop ) ) ) ) ), ...
                outDim );
% 3/27/2018 Amap = 1
%               UtilTK.expand2x2( obj.Amap .* inField, Nprop ) ) ) ) ), ...
%                outDim );
            
            if scaleN ~= 1
                outField = scaleN * outField;
            end
            
            if obj.ratio < 0
                % rot90( field, 2 ) because axis is reversed
                % - because, after the propagation, 1/ratio is multiplied, ratio is signed
                % 1/abs(ratio) = Wfft(in) / Wfft(out), this part is taken care by the normalization
                % but the sign is not. So it is absorbed here.
                outField = -rot90(outField,2);
            end
            
            % add phase change due to sideband frequency and cavity length offset
            % delL * kSB ~ 10^-9 is neglected
            outField = outField * exp( -1i * ( obj.ifo.fields.kCR * obj.delL + kSB * obj.propL0 ) );
        end
        
        %%
        function [outField, xl, yl] = propOnDemand( obj, Wout, Lprop )
            % let the field propagate by Lprop, with the output window size Wout
            
            % this is for book keeping, no calculation in the constructor
            % L0 is set to be this, not Lprop, to calculate nonimal window size
            tempProp = PropObj( 'tempProp', obj.L0, 0, 0, 1, obj.inPortID, obj.outPortID, obj.ifo );
            % calculate Amap and Pmap for the propagation calculation
            Wout = tempProp.setupMisc( Wout, Lprop );
            % this is for SB propagation, and Lprop = 0 case for CR
            tempProp.propL0 = Lprop;
            
            % calculate field propagation and keep the output large
            % the phiCorr is different when the propagator condition is changed. 
            % So correct the difference before propagates
            % phiCorr of the source mirror = -obj.CLa * obj.ratio; % (ratio - 1)/L
            % phiCorr of the sink mirror   = -obj.CLa;
            % use conv2stdSTATIC to convert once to standard form by -obj.CLa * obj.ratio, then back to internal by -(-tempProp.CLa * tempProp.ratio)
            % just phase correction, so use Wfft = 1
            xIn = obj.ifo.fields.getXlist(obj.inPortID);
            [xIn, yIn] = meshgrid( xIn, xIn );
            outField = tempProp.propField( ...
                FieldObj.conv2stdSTATIC( obj.ifo.fields.fieldList(obj.inPortID).EFFT, -obj.CLa * obj.ratio + tempProp.CLa * tempProp.ratio, xIn.^2+yIn.^2, 1, obj.ifo.fields.kCR ), ...
                obj.ifo.fields.kSB, true );
            
            % output field coordinate is, size of Wout and N = size(outField,2)
            Nout = size(outField,2);
            xl = Wout / Nout * (-Nout/2+0.5:Nout/2-0.5);
            yl = xl;
            [x,y] = meshgrid( xl, yl );
            
            % phiCorr for outField = -CLa = (1 - 1/thisRatio) / Lprop
            outField = FieldObj.conv2stdSTATIC( outField, -tempProp.CLa, x.^2+y.^2, Wout*FFTTK.aliasMargin, obj.ifo.fields.kCR );
        end
        
        %%
        function loss = propLoss( obj )
            % 1 - power in the sink optic / power at the source
            if obj.L0 == 0
                loss = 0;
            else
                optID = obj.ifo.optics.findOptic( obj.outPortID );
                
                obj.ifo.fields.calcIfNecessary( obj.inPortID );
                obj.ifo.fields.calcIfNecessary( obj.outPortID );
                
                Aperture = obj.ifo.optics.optList(optID).Aperture;
                loss = 1 - obj.ifo.fields.fieldList(obj.outPortID).calcSumIn(Aperture/2)  / obj.ifo.fields.fieldList(obj.inPortID).calcPower;
            end
        end
        
        function done = resetDelL( obj, phase )
            % set the delL of the cavity to cancel the gouy phase, and the total propagator phase becomes phase
            % resonant cond for one prop is -k l n + eta00 = n pi => l = ( mod(eta00, pi) - phase ) / (k n)
            % done = true when both ends are connected
            done = obj.inPortID > 0 && obj.outPortID > 0;
            if done
                if ~exist( 'phase', 'var' )
                    phase = obj.lockPhase;
                end
                obj.delL = ( mod( obj.gouy00 * obj.ifo.optics.getGouyMul, pi) - phase ) / (obj.ifo.fields.kCR * obj.nRef);
            end
        end
        
        function set.L0( obj, L0 )
            obj.L0 = round( L0 / obj.ifo.fields.lamCR ) * obj.ifo.fields.lamCR;
            obj.updateNdep;
        end
        
        function updateNdep( obj, ~, ~ )
            obj.propABCD = [ 1, obj.L0/obj.nRef; 0, 1 ];
        end
        
        %
        % function phase = kLGouyPhase( obj, traceOn )
        %
        % -kL + gouy = -(kSB * propL0 + kCR * delL + kCR * propL0 ) + gouy 
        %            =  -(kSB * propL0 + kCR * delL ) + (gouy - kCR * propL0)
        % kCR * propL0 = 2 pi propL0/lambda = 2 pi N
        % kCR * delL0 = gouy00
        % -kL + gouy = -(kSB * propL0 + kCR * (delL-delL0) ) + (gouy - gouy00)
        %
        function phase = kLGouyPhase( obj, traceOn )
            phase = -( obj.ifo.fields.kCR * obj.delL + obj.ifo.fields.kSB * obj.propL0 ) + obj.ifo.optics.getGouyMul * obj.gouy00;
            if exist('traceOn', 'var') && traceOn
                Cmn = obj.ifo.fields.modeCoef( obj.HR2HRProp );
                C2C1 = Cmn(2)/Cmn(1);
                phi21 = atan2(imag(C2C1), real(C2C1));
                if phi21 > pi/2
                    str21 = sprintf('pi - %e', pi - phi21);
                elseif phi21 < -pi/2
                    str21 = sprintf('-pi + %e', phi21 + pi);
                else
                    str21 = sprintf('%e', phi21 );
                end
                fprintf(1, [' > length,phase of prop \"%s\" :\n', ...
                    '        kCR = %e, kSB = %e, propL0 = %e, delL = %e, gouy00 = %e\n', ...
                    '        Gouy-kL = %e, phase(FFT) = %e EXP(i %e)/%e EXP(i %e) => %s\n'] , ...
                    obj.name, obj.ifo.fields.kCR, obj.ifo.fields.kSB, obj.propL0, obj.delL, obj.gouy00, ...
                    phase, ...
                    abs(Cmn(2)), atan2(imag(Cmn(2)), real(Cmn(2))), abs(Cmn(1)), atan2(imag(Cmn(1)), real(Cmn(1))), ...
                    str21 );
            end
        end
            
        % function [inPort, outPort] = HR2HRProp( obj )
        %
        % calculate the change of 00 mode coefficient, always between HR and HR
        % For HR/AR on mirror 1 -> AR on mirror 2, calculate HR/AR on mirror 1 to HR on mirror 2
        % For AR on mirror 1 -> HR/AR on mirror 2, calculate HR on mirror 1 to HR/AR on mirror 2
        function CinOut = HR2HRProp( obj )
            [optID, portIndex] = obj.ifo.optics.findOptic( obj.inPortID );
            % if it is AR side out , find HR side port
            if ~obj.ifo.optics.optList(optID).isHRsideIndex( portIndex ) && portIndex > 0
                [CinOut(1), ~] = obj.ifo.optics.findInOutPort( obj.inPortID );
            else
                CinOut(1) = obj.inPortID;
            end
            [optID, portIndex] = obj.ifo.optics.findOptic( obj.outPortID );
            % if it is AR side input, find HR side port
            if ~obj.ifo.optics.optList(optID).isHRsideIndex( portIndex ) && portIndex < 0
                [~, CinOut(2)] = obj.ifo.optics.findInOutPort( obj.outPortID );
                else
                    CinOut(2) = obj.outPortID;
            end
        end
        
        function info = printInfo( obj )
            info = sprintf('name: %s   prop loss = %g\n', obj.name, obj.propLoss );
            info = sprintf('%sProp from ''%s''[%d] to ''%s''[%d]\n', info, ...
                obj.ifo.fields.getName(obj.inPortID), obj.inPortID, obj.ifo.fields.getName(obj.outPortID), obj.outPortID );
            info = sprintf('%sL0 = %g, propL0 = %g, delL = %g, lockPhase/pi = %g, gouy00 = %g\n', ...
                info, obj.L0, obj.propL0, obj.delL, obj.lockPhase/pi, obj.gouy00 );
            info = sprintf('%snRef = %g, ratio = %g, CLa = %g, (-k*L+(1+n+m)gouy00)/pi = %g\n', ...
                info, obj.nRef, obj.ratio, obj.CLa, obj.kLGouyPhase/pi );
            info = sprintf( '%spropABCD = [%10.5g, %10.5g; %10.5g %10.5g]\n', info, obj.propABCD([1,3,2,4]));
        end
        
    end
    
    methods (Static)
        % was R_FFTFiltC, copied from FOG
        % cut larger angle scattering to suppress alias
        % while keeping the scattering from a bottom end of mirror to the other mirror
        function Filter = FFTFiltC(wind_length1,wind_length2, L0, Nfft, lambda)
            rhom = Nfft / (2*wind_length1);
            rhoc = wind_length2/(2*L0*lambda);
            Nc = Nfft*rhoc/rhom;
            if(Nc>Nfft)
                Nc=Nfft;
            end
            mask=ones(Nfft,Nfft);
            val = ceil((Nfft-Nc)/2);
            mask(val+1:end-val,val+1:end-val)=0;
            Filter = mask==1;
        end
    end
    
end
