function nmList = decodeNMList( nmList )
% function nmList = decodeNMList( nmList )
%   convert nmList (Wyko index or zernike coef) to a two dimentional array of zernike coefficients
%
% nmList >= 0 specifies Wyko index
% nmList < 0 specifies all Zernike terms up to n = abs(numList)
% nmList [ -nmMin, -nmMax ] sets the nmList to include n+m=nmMin to n+m=nmMax
%
    if isnumeric( nmList )
        nmin = 0;
        if length( nmList ) == 2 && nmList(1) <= 0 && nmList(2) < 0
            % like [-2,-10] reprrsents n+m = 2 to n+m = 10
            nmin = min( abs(nmList(1)), abs(nmList(2)) );
            nmList = -max( abs(nmList(1)), abs(nmList(2)) );
        end
        if length( nmList ) == 1
            nmList = int16(nmList);
            if nmList >= 0
                % this is interpreted as WykoIndex
                if nmList > 36
                    fprintf(1, 'Wyko index max 36 is used\n');
                    nmList = 36;
                end
                Wyko = nmList;
                nmList = [ 0,0; 1,1; 1,-1; 2,0; 2,2; 2,-2; 3,1; 3,-1; 4,0; 3,3; 3,-3; ...
                    4,2; 4,-2; 5,1; 5,-1; 6,0; 4,4; 4,-4; 5,3; 5,-3; 6,2; 6,-2; ...
                    7,1; 7,-1; 8,0; 5,5; 5,-5; 6,4; 6,-4; 7,3; 7,-3; 8,2; 8,-2; ...
                    9,1; 9,-1; 10,0; 12,0];
                nmList = nmList(1:Wyko+1,1:2);
            else
                nmax = abs(nmList)+1;
                nmList = zeros((nmax+1)*nmax/2 - (nmin+1)*nmin/2,2);
                ind = 0;
                for in = nmin+1:nmax
                    for im = 1:in
                        ind = ind + 1;
                        nmList(ind,1:2) = [in-1, in-2*(im-1)-1];
                    end
                end
            end
        end
    end
end