%
% fieldManager
%
%   HG coefficient at fldID port, for a standard field Fld
%     Cmn = FHGCoef( obj, m, n, fldID, Fld )
%
%   Convert internal field to standard format
%     Estd = conv2std( obj, fieldID, Eint )
%
%   set the field at port fldID in the intenal format
%     setHGField( obj, fldID, m, n, Power = 1, invQ = internal value )
%
%   apply propagation by requesting the output window size with variable propagation
%     [outField, xl, yl] = getField( fieldID, Wout, Lprop )
%
% May 24, 2020
%   LGCoef and HGCoef pass varargin properly as varargin{:} when calling modeCoef
%
% March 31, 2020
%   [outField, xl, yl] = getField( obj, fieldID, Wout, Lprop )
%       this works even when fieldID is a source and Lprop is not given.
%       Lprop is assumed to be 0
%
% November 5, 2019
%   1) many modeCoef(,,,) functions are changed to modeCoef(,,,varargin), where varargin can be (thetaX, thetaY, EFFT)
%   so that the if exist('var') is done only at the last function which actual uses these variable.
%   which are FieldObj.LGCoef and FieldObj.HGCoef
%   2) sumOut(r) is replaced by sumIn(-r)
%   
%
% Sep. 24, 2019
%   lossOutMode uses sumOut, which uses SqInCircle.
%   modeCoef( ,,, ind2, ) : default value of ind2 is 0
%
% May 2, 2019
%   findNamedField ( obj, name, check )
%      by using UtilTK.wildMatch( {obj.fieldList(:).name}, name )
%      field names are wildcard serached using name as a pattern
%   conv2FieldIDList and printSome have been changed to utilize this change
%
% April 30, 2019
%   printSome( fldIDs ) prints information specified by fldIDs
%
% April 3rd, 2019
%   SetDefField does nothing to do with dxyz, a.ka.c0, all are done in LG and HG functions.
%
% March 5, 2019
%   Cmn = modeCoef( obj, fldIDs, fieldMode, ind1, ind2, EFFT ), HGCoef, LGCoef
%      if ind1 < 0, all modes with ind1+ind2 == abs(ind1) or 2*ind1+ind2 = abs(ind1);
%      Cmn is id array
%
% March 1, 2019
%   [E, xList, C] = fieldAtFID( obj, FID, fieldMode, ind1, ind2, stdForm, invQ )
%   C is the coefficient of the mode using modeCoef
%
% January 30, 2019
%   power = calcPower( fieldID, isSrc )
%   When isSrc is true, the power is calculated using the soruce of the field.
%   This is better for far propagation with spreading field, like PRM2 to PRM3
%
% August 24, 2018
%   lossOutMode is protected for the case of out-going field, 
%   in case the reflection is not a good replacement, use transmission
%
% July 30, 2018
%   fieldAtFID has invQ as the last optiona argument.
%
% July 19, 2018
%   modeSum( , mnmax ) sums only n+m = abs(mnmax) if mnmax < 0.
%
% July 7, 2018
%   function [loss, srcPower] = lossOutMode( obj, fieldID, mode, index1, index2 ) added
%   This calculates the loss of fieldID due to the optics size
%
% May 24, 2018
%   function [Esum, x] = modeAmpSum( obj, fldID, modes, fMode ) added
%   modes is data returned by FFTIFO.mainModes, etc
%
% May 18, 2018
%   new function added : fieldList = conv2FieldIDList( obj, fList, check )
%   This converts multiple field specifation to an array of field IDs, replace findNamedField
%
% April 7, 2018
%   calcIfNecessary and its use is simplified. findNamedField does not call calcIfNecessary any more.
%
% March 9, 2018
%   function Cmn = modeCoef( obj, fldIDs, fieldMode, ind1, ind2 ) added
%      LGCoef and HGCoef combined. If fieldMode is not specified, laser mode is used
%   LGCoef and HGCoef do not accept amplitude E as an extra argument
%
% December 4th, 2017
%   In fieldAtFID and fieldAtFIDMode, findNamedField is called with default check, i.e., the field is calculated if necessary.
%
% December 1st, 2017
%   [modeEs, idList, Clist, Etotal, xList] = fieldAtFIDMode( FID, fieldMode, ind1, ind2, threshold )
%   this returns modes with power fracton larger than therhold, 
%   e.g., fieldAtFIDMode( 'BS-ARy-o, 'LG', 0:5, 0, 0.1 )
%   returns LG(id1,id2) whose power fraction are larger than 10%
%   idList(:,1:2) returns a list of (id1,id2), Clist(:,1,2) 
%   returns the coefficient and power fraction of LG(id1,id2) mode
%   modeEs(id,:,:) is the amplitude of LG(id1,id2)
%
% September 8th, 2017
%   setDefField( obj, fldID, Power, invQ ) can have an optional last argument invQ
%
% April 21, 2017
%   setLaserField is added, which calls laserObj to set the laser source field, which could be of non-gaussian shape.
%   function setDefField uses conv2int now.
%
% April 15, 2017
%   getField is updated to support Wout and Lprop properly. See PropObj for more details.
%
% April 1, 2017
%   fieldAtFID has a new argument, stdForm. If this is false, E is returned in internal form.
%
% March 2, 2017
%   setDefField( obj, fldID, Power, invQ ) is modified that,
%   if Power <= 0, obj.Eamp^2 is used as the power.
%   If this set the power to be 0, it is set to 1.
%
% September 20, 2016
%   Sometimes, fields with status = 0 is not properly recalculated.
%   For a PRFPM case, when an arm is removed, ITM-HR-o is not properly updated and recalculated. ???
%
% May 26, 2016
%   getField( fieldID, W, Lprop ) uses fieldID as the sink starting from source, or fieldID as the source, when Lprop > 0.
%   getField( 'ITM-HR-i', W, 2000 ) and getField( 'ETM-HR-o', W, 2000 ) returns the same answer.
%
% May 13, 2016
%   setDefField( obj, fldID, Power, invQ )
%      set the default mode field (specified at addLaser) at port fldID in the internal format
%      originally, correction by phiCorr was included by using invRoC - phiCorr when calling HG/LG function
%      Now, to support tilt and off centering, this is changed to pass absolute value of RoC
%
% October 5, 2015
%   getPointScatter is moved to a separate file, getPointScatter.m
%   [dEsum, E0, xl, dPsum] = getPointScatter( fields, srcID, defects, WfftOut, L )
%
% September 19th, 2015
%   circleAnom and squareAnom completed which are used in getPointScatter
%   to integrate areas in a cicle and square
%
% September 18th, 2015
%   * [dE, E0, xl] = getPointScatter( obj, srcID, defects, WfftOut, L )
%   modified to include multiple share shaped defects specified by locations and size.
%   * getField modified so that, when Lprop is the same as an existing
%   propagator, don't prop again, just copy existing data, like Lprop = 0 case
%
% September 2nd, 2015
%   lam0 and kLam0 are stored in fieldManager object, moved from ConstantTK
%   This is modified by addLaser
%
% October 27, 2014
%   conv2std is changed to return zero matrix when the field is not ready
%
% October 26, 2014
%   handle of FFTIFO itself is saved in all manager objects,
%   instead of handle of multiple manager handles
%
% August 29, 2014
%   function [Fld, xl, yl] = getField( fldID, Wout, Lprop ) <== used to be onDemand
%   getField has a new argument, Lprop.
%
% July 31, 2014
%   1) input beam can be LGpm or HGmn
%      setInputType( type, ind1, ind2 ) is the interface to define these
%   2) [E, xList] = fieldAtFID( fieldID, 'LG or HG', p/m, m/n ) added to get field of specified type
%      fieldAtFID(fieldID, no field specification) returns using the type specified at addLaser
%
%
% July 28, 2014
%   findNamedField ( obj, name, check ) when check < 0, calcIfNecessary is not called

classdef fieldManager < handle
    
    properties
        ifo
        
        fieldList
        
        % CR field specification. These two need to be set properly.
        % aetherically better to use set/get, but, for now, these are set
        % in addLaser by hand
        lamCR
        kCR
        
        % sideband wavelength
        kSB
    end
    
    methods
        
        function obj = fieldManager( ifo )
            obj.ifo = ifo;
            
            obj.fieldList = FieldObj.empty(0);
            
            % CR frequency
            obj.kSB = zeros(1,1);
            
            obj.lamCR = nan;
            obj.kCR = nan;
        end
        
        function xList = getXlist( obj, fieldID )
            xList = obj.fieldList(fieldID).getXlist;
        end
        
        function Wfft = getWfft( obj, fldID )
            Wfft = obj.fieldList(fldID).getWfft;
        end
        
        function name = getName( obj, ind )
            if ind > 0 && ind <= length( obj.fieldList )
                name = obj.fieldList(ind).name;
            else
                name = 'unknown';
            end
        end
        
        function qInv = getQInv( obj, fldID, reverseDir )            
            fldID = obj.findNamedField( fldID, -1 );
            qInv = obj.fieldList(fldID).invQ;
            if exist( 'reverseDir', 'var' ) && reverseDir
                qInv = MMTK.reverseDir( qInv );
            end
        end
        
        % be careful, qInv updates the following quantities
        %   obj.w = MMTK.qinv2W( newInvQ, obj.ifo.fields.kCR );
        %   obj.invRoC = real( newInvQ );
        %   obj.z = real( 1/newInvQ );
        %   obj.z0 = imag( 1/newInvQ );
        function setQInv( obj, fldID, qInv, reverseDir )
            fldID = obj.findNamedField( fldID, -1 );
            if exist( 'reverseDir', 'var' ) && reverseDir
                qInv = MMTK.reverseDir( qInv );
            end
            obj.fieldList(fldID).invQ = qInv;
        end
        
        function setLaserField( obj )
            % setLaserField( obj ) 
            % set the FFT amplitude of the laser field
            
            % find the laser ID
            [portID, laserID] = obj.ifo.optics.findLaser;
            
            % if settings are not ready, stop here
            % this protection for the case that FFTIFO.setInputType is called in the defineIFO code
            if isnan( obj.fieldList(portID).invQ )
                return;
            end
            
            % fill the inpit field using the default mode
            obj.ifo.optics.optList(laserID).setDefField(portID);
        end       
        
        % field (internal format) and field ID of the laser transmitted or reflected at the first mirror interacted
        % if fldID is not specified, transmitted field is returned
        % This is used to calculate the external source injected to IFO
        function [EFFT, fldID] = laserInIFO( obj, fldID )
            srcID = obj.ifo.optics.findLaser;
            % sinkID is input to the first optic hit by laser
            [sinkID, propID] = obj.ifo.props.getPropTo( srcID, true );
            if ~exist( 'fldID', 'var' )
                fldID = 0;
                EFFT = 0;
            end
            if sinkID > 0
                obj.ifo.props.propField(propID);
                transPort = obj.ifo.optics.findTransPort( sinkID );
                reflPort = obj.ifo.optics.findReflPort( sinkID );
                if fldID == 0
                    if transPort > 0
                        fldID = transPort;
                    elseif reflPort > 0
                        fldID = reflPort;
                    end
                end
                
                if fldID == transPort
                    EFFT = obj.ifo.optics.trans( sinkID );
                elseif fldID == reflPort
                    EFFT = obj.ifo.optics.refl( sinkID );
                end
                
            end
        end
        
        function [modeEs, idList, Clist, Etotal, xList] = fieldAtFIDMode( obj, FID, fieldMode, ind1, ind2, threshold )
            %
            % [Es, xList, Cs] = fieldAtFIDMode( obj, FID, fieldMode, ind1, ind2, threshold )
            % returns mode amplitudes whose power fraction is larger than threshold
            % Example
            % inputs
            %   FID = 'BS-ARy-o' ; dark port output
            %   fieldMode = 'LG'; LG mode, can be 'HG'
            %   ind1 = 0:5, ind2 = 0; check LG(ind1,ind2)
            %   threshould = 0.1; pickup LG modes whose power fraction is larger than 10% of total energy
            % outputs
            %   Etota;(:,:) total field at this port
            %   modeEs(:,:, id) : id'th field 
            %   xList : x (and y) coordinate of fields, modeEs(id,iy,ix) is field at (x,y) = xList(ix,iy)
            %   idList(1:2, id) : modeEs are amplitudes of LG( idList(1,id), idList(2,id) ) whose power fraction > threshold
            %   Clist(1:2, id) : Amplitude (1) and power fraction (2) of LG( idList(1,id), idList(2,id) )
            %
            FID = obj.findNamedField( FID );
            [Etotal, xList] = obj.getField( FID, 0 );
            Ptotal = obj.calcPower( FID );
            idList = zeros(2, 0);
            Clist = zeros(2, 0);
            modeEs = zeros( [size( Etotal ), 0] );
            for id2 = 1:length(ind2)
                for id1 = 1:length(ind1)
                    if strcmpi(fieldMode, 'LG')
                        C = obj.LGCoef(FID, ind1(id1), ind2(id2));
                    else
                        C = obj.HGCoef(FID, ind1(id1), ind2(id2));
                    end
                    frac = abs(C)^2 / Ptotal;
                    if frac >= threshold
                        id = size(Clist,2)+1;
                        Clist(:,id) = [C; frac];
                        idList(:, id) = [ind1(id1); ind2(id2)];
                        modeEs(:,:, id) = C * obj.fieldAtFID(FID,fieldMode,ind1(id1),ind2(id2));
                    end
                end
            end
        end                        

        function [E, xList, C] = fieldAtFID( obj, FID, fieldMode, ind1, ind2, stdForm, invQ )
            % function [E, xList] = fieldAtFID( fieldID, 'LG or HG', p/m, m/n, stdForm )
            % This returns 1W field of the type (LG or HG) in standard form.
            % If called without any arguments, the field type specified at addLaser is used
            % If stdForm is explicitly specified to be false, return internal format
            
            FID = obj.findNamedField( FID );
            [fieldModeDef, ind1Def, ind2Def] = obj.ifo.optics.getInputType;
            if ~exist('fieldMode', 'var')
                fieldMode = fieldModeDef;
            end
            if ~exist('ind1', 'var')
                ind1 = ind1Def;
            end
            if ~exist('ind2', 'var')
                ind2 = ind2Def;
            end
            
            if ~exist( 'invQ', 'var' )
                invQ = obj.fieldList(FID).invQ;
            end
            if strcmp( fieldMode, 'LG')
                [E, xList] = obj.fieldList(FID).LGFieldatFID( ind1, ind2, invQ );
            else
                [E, xList] = obj.fieldList(FID).HGField2DatFID( ind1, ind2, invQ );
            end
            
            if nargout == 3
                C = obj.modeCoef( FID, fieldMode, ind1, ind2 );
            end
            
            if exist( 'stdForm', 'var' ) && ~stdForm
                E = obj.fieldList(FID).conv2int( E );
            end
        end
        
        function markFieldStatus( obj, fieldID, guessID )
            obj.fieldList(fieldID).status = guessID;
        end
        
        % update fields with status == 0
        % fList can be 'ALL', -1, or '' for all fields
        function updated = calcIfNecessary( obj, fList )
            updated = false;
            if ~exist('fList', 'var')
                fList = '';
            end
            fList = obj.conv2FieldIDList( fList );
            
            for id = 1 : length(fList)
                fID = fList(id);
                % if current status == 0, i.e., no field yet ...
                update = obj.fieldList(fID).status == 0;
                if update
                    if ~obj.ifo.cavities.loopFieldReady || isnan( obj.fieldList(fID).invQ )
                        return;
                    end
                    obj.fieldList(fID).updateField;
                    updated = true;
                end
            end
        end
        
        % if isSrc is true and if this fieldID has a source, use that field
        % E.g., fieldID is 'ETM-HR-i' and isSrc is true,
        % fieldID is 'ITM-HR-o' which propagates to ETM
        % field ETM-HR-i may not have full power, lost of out of FFT window size
        function power = calcPower( obj, fieldID, isSrc )
            fieldID = obj.findNamedField( fieldID, -1 );
            if exist('isSrc', 'var') && isSrc 
                srcID = obj.ifo.props.getPropFrom( fieldID );
                if srcID > 0
                    fieldID = srcID;
                end
            end
            power = obj.fieldList(fieldID).calcPower;
        end
        
%       function [loss, srcPower] = lossOutMode( obj, fieldID, mode, index1, index2 )
%            fieldID is on the sink mirror (like ETM), mode = LG(index1,index2) or 'HG(index1,index2)
%            this mode propagates from the srsOptic (line ITM) to the sinkOptic (like ETM)
%            loss is 1 - power on the sinkMirror / power on the srcMirror of this mode
        function [loss, srcAmp] = lossOutMode( obj, fieldID, fieldMode, ind1, ind2 )
            fieldID = obj.findNamedField( fieldID, -1 );
            if obj.ifo.fields.fieldList(fieldID).portIDfrom > 0
                % in this function, this is not so important
                fieldID = obj.ifo.fields.fieldList(fieldID).portIDfrom;
            end
            opticID = obj.ifo.optics.findOptic( fieldID, true );            
            fromPortID = obj.ifo.props.getPropFrom( fieldID );
            % if fieldID is output field, find the incoming field which is reflected to fieldID
            if fromPortID < 0
                fieldIDX = obj.ifo.optics.findReflPort( fieldID, true );
                fromPortID = obj.ifo.props.getPropFrom( fieldIDX );
                % if the field by findReflPort is not good, use findTrasPort
                if fromPortID < 0
                    fieldIDX = obj.ifo.optics.findTransPort( fieldID, true );
                    fromPortID = obj.ifo.props.getPropFrom( fieldIDX );
                end
            end
            srcAmp = obj.ifo.modeCoef( fromPortID, fieldMode, ind1, ind2 );
            [E, xList] = obj.ifo.fields.fieldAtFID( fieldID, fieldMode, ind1, ind2 );
            loss = UtilTK.sumIn( xList, E, -obj.ifo.optics.optList(opticID).Aperture/2 );
        end
        
        function loss = lossOut( obj, fieldID )
            % field power out of a circle optic
            %    loss = 1 - power in aperture / total power in
            fieldID = obj.findNamedField( fieldID, -1 );
            fromPortID = obj.ifo.props.getPropFrom( fieldID );
            opticID = obj.ifo.optics.findOptic( fieldID );
            loss = obj.fieldList(fieldID).calcSumIn(-obj.ifo.optics.optList(opticID).Aperture/2) / obj.calcPower(fromPortID);
        end
        
        function [outField, xl, yl] = getField( obj, fieldID, Wout, Lprop )
            % apply propagation with specifictaions of the output window size and propagatio distance
            %
            % [Fld, x, y] = getField( obj, fldID, Wout, Lprop )
            % return the 2D field map (Fld) and coordinates (x, y) of the field fldID.
            % Fld(jy,ix) is the complex amplitude at x(ix), y(jy) in units of sqrt(W)/m.
            %
            % If Wout > 0 is specified, the field is calculated in (Wout x Wout) area
            % If Wout = 0, Wfft is set to be optimal based on the beam size
            % If Wout = -1, Wfft is set to be the same as the sink.
            %
            % If Lprop is specified, the field is calculated Lprop from the HR side of the source.
            % If Lprop is specified and if fieldID is a source of a propagation,
            % it returns the same result when the fieldID is the sink of the propagation.
            % I.e., getField('X-HR-o', w, prop) = getField('Y-HR-i', w, prop ) when X-HR-o is connected to Y-HR-i.
            % If Lprop is not specified, Wout ~= 0 and fieldID is a spurce of a prop, Lprop = 0, i.e., it is the field at the sorce
            
            [srcID, propID] = obj.ifo.props.getPropFrom( fieldID );
            
             % fieldID is not an end of propagation, so check if it is start of prop
            if srcID < 0
                [sinkID, propID] = obj.ifo.props.getPropTo( fieldID );
                if sinkID <= 0
                    error( 'this fieldID (%d) is not valid.\n', fieldID );
                else
                    fieldID = sinkID;
                    if ~exist( 'Lprop', 'var' ) || isnan( Lprop )
                        Lprop = 0;
                    end
                end
            end

            if ~exist( 'Lprop', 'var' ) || propID <= 0
                % normal propagation distance
                Lprop = nan;                
            elseif ~isnan( Lprop )
                % now the field propagation done by propID, and the end is fiedID
                if abs(obj.ifo.props.getPropLeng(propID) / Lprop - 1) < 1e-10
                    % if Lprop is the propator prop length, within numerical accuracy, say to use default
                    Lprop = nan;
                end
            end
            
            if isnan( Lprop )
                % propagation distance is the default
                if (Wout <= obj.getWfft(fieldID)/2) && (Wout >= 0)
                    % if Wout is small, no need to expand
                    outField = obj.conv2std( fieldID );
                    xl = obj.getXlist(fieldID);
                    yl = xl;
                    return;
                else
                    Lprop = obj.ifo.props.getPropLeng(propID);
                end
            end
            
            % Lprop = 0 case is included in propOnDemand
            % we need some change of the field
            % Now, Lprop is positive finite
            % If Wout > 0 is specified, the field is calculated in (Wout x Wout) area
            % If Wout = 0, Wfft is set to be optimal based on the beam size
            % If Wout = -1, Wfft is set to be the same as the sink of the nominal propagation.
            [outField, xl, yl] = ...
                obj.ifo.props.propList(propID).propOnDemand( Wout, Lprop );
            
        end
        
        function clearFFTFields( obj, fldIDs )
            if ~exist( 'fldID', 'var' )
                fldIDs = 1:length( obj.fieldList);
            end
            for id = fldIDs
                obj.fieldList(id).EFFT = [];
            end
        end
        
        %
        % fidID is used to find the q value of the field
        % modes(imn,1:3) = index1, index2, amplitude
        % output = sum( HG/LG(ind1,ind2, qVal) * amplitude
        % fModes = 'HG' or 'LG', default is 'HG'
        %
        % if modes(:,3) does not exist or nan, use modeCoef
        % if modes is a number, it means all modes with
        %    if modes < 0, all modes with n+m = abs(modes)
        %    if modes >=0, all modes with n+m <= modes
        %
        function [Esum, x] = modeAmpSum( obj, fldID, modes, fMode )
            fldID = obj.conv2FieldIDList(fldID, -1);
            x = obj.getXlist(fldID);
            Esum = zeros( length(x), length(x) );
            if ~exist( 'fMode', 'var' )
                fMode = 'HG';
            end
            isLG = strcmpi(fMode, 'LG');
            
            if length( modes ) == 1
                if isLG
                    modes = LGTK.LGModes( modes );
                else
                    modes = HGTK.HGModes( modes );
                end
            end

            for id = 1 : size(modes,1)
                if size(modes,2) > 2 && ~isnan( modes(id,3) )
                    C = modes(id,3);
                else
                    C = obj.modeCoef( fldID, fMode, modes(id,1), modes(id,2) );
                end
                if isLG
                    Esum = Esum + obj.fieldList(fldID).LGFieldatFID( modes(id,1), modes(id, 2) ) * C;
                else
                    Esum = Esum + obj.fieldList(fldID).HGField2DatFID( modes(id,1), modes(id, 2) ) * C;
                end
            end
        end
        
        % coefficients of mode amplitudes
        % if fieldMode or ind1 or ind2 are missing, input laser mode is used
        % if ind1 < 0, all modes with ind1+ind2 == abs(ind1) or 2*ind1+ind2 = abs(ind1);  
        % varargin = (thetaX, thetaY, EFFT)
        function Cmn = modeCoef( obj, fldIDs, fieldMode, ind1, ind2, varargin )
            fldIDs = obj.conv2FieldIDList(fldIDs, -1);
            if ~exist( 'fieldMode', 'var' ) || isempty( fieldMode )
                [fieldMode, ind1, ind2] = obj.ifo.optics.getInputType;
            end
                        
            isLG = strcmpi(fieldMode, 'LG');
            oneModes = ind1 < 0;
            if oneModes
                if isLG
                    modeList = LGTK.LGModes( ind1 );
                else
                    modeList = HGTK.HGModes( ind1 );
                end
                % LG(-3) : (1,1),(1,-1),(0,3),(0,-3)
                % HG(-3) : (0,3),(1,2),(2,1),(3,0)
                ind1 = modeList(:,1);
                ind2 = modeList(:,2);
                Cmn = zeros( length(fldIDs), length(ind1) );
            else
                if ~exist( 'ind2', 'var' )
                    ind2 = 0;
                end
                Cmn = zeros( length(fldIDs), length(ind1), length(ind2) );
            end
            
            for iFld = 1 : length(fldIDs)
                if oneModes  
                    % coef of fields whose indexes are (ind1(1),ind2(1)), (ind1(2),ind2(2)), ...
                    for inm = 1 : length( ind1 )
                        if isLG
                            Cmn(iFld, inm) = obj.fieldList(fldIDs(iFld)).LGCoef(ind1(inm), ind2(inm), varargin{:});
                        else
                            Cmn(iFld, inm) = obj.fieldList(fldIDs(iFld)).HGCoef(ind1(inm), ind2(inm), varargin{:});
                        end
                    end
                else                    
                    % coef of fields whose indexes are 
                    % (ind1(1),ind2(1)), (ind1(2),ind2(1)), ...(ind1(end),ind2(1)), 
                    % (ind1(1),ind2(2)), (ind1(2),ind2(2)), ...(ind1(end),ind2(2)),
                    % ,,,
                    % (ind1(1),ind2(end)), ... (ind1(end),ind2(end))
                    if isLG
                        Cmn(iFld, :, :) = obj.fieldList(fldIDs(iFld)).LGCoef(ind1, ind2, varargin{:});
                    else
                        Cmn(iFld, :, :) = obj.fieldList(fldIDs(iFld)).HGCoef(ind1, ind2, varargin{:});
                    end
                end
            end
        end
        
        %   HG coefficient at fldID port, for a standard field Fld
        function Cmn = HGCoef( obj, fldIDs, m, n, varargin )
            Cmn = obj.modeCoef( fldIDs, 'HG', m, n, varargin{:} );
        end
        
        %   LG coefficient at fldID port, for a standard field Fld
        function Cpm = LGCoef( obj, fldIDs, p, m, varargin )
            Cpm = obj.modeCoef( fldIDs, 'LG', p, m, varargin{:} );
        end
        
        % sum of HG(m,n) powers with m+n < mnmax
        % if mnmax < 0, it is the sum of only |mnmax|, i.e.,
        % for mnmax = 3, this is a sum of (3,0),(2,1),(1,2),(0,3),(2,0),(1,1),(0,2),(1,0),(0,1),(0,0)
        % for mnmax = -3, this is a sum of (3,0),(2,1),(1,2),(0,3)
        function pwrMN = modeSum( obj, fldID, mnmax )
            pwrMN = zeros(1, length(mnmax));
            for imn = 1 : length(mnmax)
                % sum goes from large modes to small modes to minimize the truncation error
                if mnmax(imn) < 0
                    mnmin = -mnmax(imn);
                else
                    mnmin = 0;
                end
                for mn = abs(mnmax(imn)) : -1 : mnmin
                    for m = 0 : mn
                        pwrMN(imn) = pwrMN(imn) + abs(obj.HGCoef( fldID, m, mn-m )).^2;
                    end
                end
            end
        end
        
        function setDefField( obj, fldID, Power, invQ )
            % set the default mode field (specified at addLaser) at port fldID in the internal format
            if ~exist( 'Power', 'var' )
                Power = 1;
            elseif Power <= 0
                Power = abs(obj.fieldList(fldID).Eamp)^2;
                if isnan( Power ) || Power == 0
                    Power = 1;
                end
            end
            % this is a field with the internal format
            %   Einternal = Estandard * exp( -i phiCorr ) * Wfft
            laserPort = obj.ifo.optics.findLaser;
            [fieldMode, ind1, ind2, dxyz, nvec] = obj.ifo.optics.getInputType;
            % use dxyz and nvec only for laser source
            if fldID ~= laserPort
                dxyz = [0,0,0];
                nvec = [0,0];
            end
            
            if ~exist( 'invQ', 'var' ) || isnan( invQ )
                invQ = obj.fieldList(fldID).invQ;
            end
            
            xList = obj.getXlist( fldID );
            [invRoC, w] = MMTK.qinv2RinvW( invQ, obj.kCR );

            % LGFieldRW and HGField2DRW need absolute value of RoC,
            % so the correction by phiCorr is explicitly calculated
            if strcmp(fieldMode, 'LG')
                obj.fieldList( fldID ).EFFT = ...
                    LGTK.LGFieldRW( ind1, ind2, xList, xList, ...
                    invRoC, w, obj.kCR, dxyz, nvec );
            else
                obj.fieldList( fldID ).EFFT = ...
                    HGTK.HGField2DRW( ind1, ind2, xList, xList, ...
                    invRoC, w, obj.kCR, dxyz, nvec );
            end
            % E(internal form) = E(standard form) * exp( +i k/2 r^2 phiCorr ) * Wfft
            obj.fieldList( fldID ).conv2int( obj.fieldList( fldID ).EFFT * sqrt(Power) );
       end
        
        function setHGField( obj, fldID, m, n, Power, invQ )
            % set the field at port fldID in the intenal format
            if ~exist( 'Power', 'var' )
                Power = 1;
            end
            if ~exist( 'invQ', 'var' )
                invQ = obj.fieldList(fldID).invQ;
            end
            % this is a field with the internal format
            %   Einternal = Estandard * exp( -i phiCorr ) * Wfft
            xList = obj.getXlist( fldID );
            [invRoC, w] = MMTK.qinv2RinvW( invQ, obj.kCR );
            obj.fieldList( fldID ).EFFT = ...
                sqrt(Power) * HGTK.HGField2DRW( m, n, xList, xList, ...
                invRoC - obj.fieldList(fldID).phiCorr, w, obj.kCR ) * obj.getWfft( fldID );
        end
        
        function invRoC = getInvRoC( obj, fldID )
            invRoC = obj.fieldList(fldID).invRoC;
        end
        
        function Estd = conv2std( obj, fieldID, Eint )
            %   Convert internal field to standard format
            if ~exist( 'Eint', 'var' )
                if isempty( obj.fieldList(fieldID).EFFT )
                    fprintf(1,'Internal field is not defined : %s\n', obj.fieldList(fieldID).name);
                    Estd = obj.ifo.ffts.zeroMat;
                    return
                end
                Eint = obj.fieldList(fieldID).EFFT;
            end
            Estd = obj.fieldList(fieldID).conv2std( Eint );
        end
        
        % create a new field.
        % If coyFromFid is specified,
        %     all data are copifed from the src field and the status is set to 0
        function id = newField( obj, name, copyFromFid )
            % make sure this does not exist
            id = find( strcmp( name, { obj.fieldList(:).name } ), 1 );
            if isempty( id )
                id = length( obj.fieldList ) + 1;
            end
            if exist( 'copyFromFid', 'var' )
                obj.fieldList(id) = copy( obj.fieldList(copyFromFid) );
                obj.fieldList(id).name = name;
            else
                obj.fieldList(id) = FieldObj( name, obj.ifo );
            end
            obj.markFieldStatus(id, 0);
        end
        
        function name = findName( obj, id )
            if (id > 0) && (id <= length(obj.fieldList))
                name = obj.fieldList(id).name;
            else
                name = 'noField';
            end
        end
        
        % fieldID, or field name or array of them are converted to array of fieldIDs
        % check >= 0 does fill uncalculated field
        % fList == 'ALL' or fList = -1 or fList == '', fList is set to be a full list
        % Modified for the case that fList can have wildcard * and findNamedField may return multiple matches
        % like 'BS-AR*o' matching with BS-ARx-o and BS-ARy-o
        function fieldList = conv2FieldIDList( obj, fList, check )
            if ~exist( 'check', 'var' )
                check = 0;
            end
            if iscell(fList)
                fieldList = zeros(0);
                for fid = 1 : length(fList)
                    fieldList = [fieldList, obj.findNamedField( fList{fid}, check )];
                end
            elseif isempty(fList) || strcmpi( fList, 'ALL' ) || fList(1) == -1
                fieldList = 1:length(obj.fieldList);
            else
                if ischar( fList ) || length( fList ) == 1
                    fieldList = obj.findNamedField( fList, check );
                else
                    fieldList = fList;
                end
            end
        end
        
        function id = findNamedField ( obj, name, check )
            % function id = findNamedField ( name, check )
            %
            % Return field id for the field named name
            % If the name is not a valid field name,
            %   if check ~= 0, return id = 0 without interuption
            %   if check == 0, execution exists by error call
            %
            % If name is valid or name is ID number,
            %   if check >= 0, calcIfNecessary is called and, if field is not ready, i.e.,
            %      status is 0, the field is calculated
            %   if check < 0, id is returned without calcIfNecessary
            % if the name is not explicit, i.e, no matching
            %    UtilTK.wildMatch( {obj.fieldList(:).name}, name )
            %    is used to find wildcard finding, like 'BS*AR*'
            if ~exist( 'check', 'var' )
                check = 0;
            end
            if ischar( name )
                id = find( strcmp( name, { obj.fieldList(:).name } ), 1 );
                if isempty( id )
                    id = find( strcmp( name, { obj.ifo.optics.optList(:).name } ), 1 );
                    if ~isempty( id )
                        if obj.ifo.optics.optList(id).isLaserObj
                            id = obj.ifo.optics.findPortID( sprintf('%s-1-o', name) );
                        elseif obj.ifo.optics.optList(id).isDetectorObj
                            id = obj.ifo.optics.findPortID( sprintf('%s-1-i', name) );
                        else
                            id = [];
                        end
                    end
                    if isempty( id )
                        id = UtilTK.wildMatch( {obj.fieldList(:).name}, name );
                        if isempty( id )
                            id = 0;
                            if ~check
                                error('Field name ''%s'' does not exist', name );
                            end
                            return;
                        end
                    end
                end
            else
                id = name;
            end
            
            % if check < 0, don't do anything
            if check >= 0
                % if the field is not ready, i.e., status == 0, calculate using available fields
                % this automated update is disabled ...
%                obj.calcIfNecessary( id );
            end
        end
        
        function setFieldName( obj, id, name )
            if (id>0) && (length(obj.fieldList)>=id)
                obj.fieldList(id).name = name;
            end
        end
        
        function phiCorr = getPhiCorr( obj, fieldID )
            phiCorr = obj.fieldList( fieldID ).phiCorr;
        end
        
        % print fields
        % if all ia false, only fields with power are printed
        function printAll( obj, fid )
            if ~exist( 'fid', 'var' )
                fid = 1;
            end
            fileOpened = ischar( fid );
            if fileOpened
                fid = fopen(fid, 'w');
            end
            
            fprintf( fid, '\n=== Information about fields ===\n\n');
            for ii = 1:length(obj.fieldList)
                info =  obj.fieldList(ii).printInfo( true );
                if ~isempty( info )
                    fprintf( fid, '%d) %s\n\n', ii, info );
                end
            end
            
            if fileOpened
                fclose(fid);
            end
        end
        
        % print fields specified by fieldIDs, like 5:10
        % fldID can be a cell string, {'abc', 'def'}, or 'abc', 'def', and a string can have a wild card, like BS*in
        function printSome( obj, varargin )
            if nargin == 1
                fldIDs = 1 : length(obj.fieldList);
            else
                fldIDs = zeros(0);
                for ii = 1:nargin-1
                    fldIDs = [fldIDs, obj.conv2FieldIDList( varargin{ii} )];
                end
            end
            
            fprintf( 1, '\n=== Information about fields ===\n\n');
            for ii = fldIDs
                info =  obj.fieldList(ii).printInfo( true );
                if ~isempty( info )
                    fprintf( 1, '%d) %s\n\n', ii, info );
                end
            end
        end
        
    end
    
    methods (Static)
    end
end
