function [f,psd] = map2PSD1D( fileName, mirrorAperture, rxy, removeZernike, mmScale )
% function [f,psd] = map2PSD1D( fileName, mirrorAperture [mirror size], rxy [0], removeZernike [3], mmScale [true] )
% PSD is calculated using data in mirrorAperture. 
% Same data is used to calculated and remove lowerOrder zernike using zernikeCoef.
% rxy = 0, 1, 2 for PSD(r) integrating over phi, PSD(x) integrating over y and PSD(y)
% If mmScale is true, unit is f in 1/mm, PSD in nm^2 mm, 
% If mmScale is false, unit is f in 1/m, PSD in m^2 m

% Uses LMA version of PSD calculator, supports windowed PSD(r)
% 

if ~exist( 'rxy', 'var' )
    rxy = 0;
end
[amap, xl, yl] = loadOneDataFile( fileName );
if ~exist( 'mirrorAperture', 'var' ) || mirrorAperture == 0
    if isnan(amap(1,1))
        ind = find( ~isnan(amap) );
    else   
        ind = find( amap ~= amap(1,1) );
    end
    [xg,yg] = meshgrid( xl, yl );
    mirrorAperture = min( max(xg(ind)) - min(xg(ind)), max(yg(ind)) - min(yg(ind)) ) * 0.99;
end

if ~exist( 'removeZernike', 'var' )
    removeZernike = 3; % up to power
end
amap = zernikeCoef( amap, xl, yl, mirrorAperture, removeZernike, true );

[f, psd] = LMA_PSD( amap, xl, yl, mirrorAperture, rxy );

if ~exist( 'mmScale', 'var' ) || mmScale
    f = f/1000;
    psd = psd * 1e21;
end

