function datOut = extendOuterside( datIn, xl, yl, rfit, rout, indMax )
% function datOut = extendOuterside( datIn, xl, yl, rfit, rout, indMax )
% fit dat in radius of rfit using zernikeCoef the index of indMax
% use the coefficient to expand data in the ring between r = rfit ~ rout

% March 6th, 2020
%   outside region was filled improperly
%
% December 17, 2014  Hiro Yamamoto
%   inital code

if ~exist( 'indMax', 'var' ) || indMax == 0
    % using smaller index seems to be better than using a large index, like 17.
    indMax = 4;
end

% calculate the zernite terms
% -indMax means : for positive indMax, use all (0,0),(1,-1),(1,1),...,(indMax,-indMax) ...,(indMax,indMax)
%               : for negative indexMax, use lower order wyko indexes from 0 to (-indMax)
% datDel = datIn - fit, 
% fit = datIn - datDel
% datOut( r < rfit ) = datIn
% datOut( r > rfit ) = fit
datDel = zernikeCoef( datIn, xl, yl, rfit*2, -indMax, -1 ); % datDel = sum of zernike terms of lower order modes

if size(xl,2) == 1
    xl = xl';
end
if size(yl,1) == 1
    yl = yl';
end
[xg, yg] = meshgrid(xl(1,:), yl(:,1));
rg = sqrt( xg.^2 + yg.^2 );

datOut = datIn;
rng = rg > rfit & rg < abs(rout);
datOut( rng ) = datDel( rng );
