function [mapL, xL] = mapDoubler( map, x, y, diam, shape )
% function [mapX, xlX, ylX] = mapDoubler( map, xl, yl, diam, shape )
% Convert data given map defined at (xl,yl) to frequency doamin using data in diam
% and expand the low frequency region to 1/(2*diam), and convert back to x domain
% (0~N/2-1)+(-N/2~-1) => (-N/2~N/2-1) => (-N~N-1) => (0~N-1)+(-N~-1)
% if shape is 'circle', data in diameter diam is used to expand to a
% circle with diameter of 2xdiam
% otherwise, data in (diam x diam) are used to generate data in (2xdiam x 2xdiam)
%
if ~exist( 'shape', 'var' )
    shape = 'c';
end

% Remove the NaN if any
map(isnan(map)) = 0;

% make nx = ny, so dx and dy should be similar
dx = x(2) - x(1);
dy = y(2) - y(1);

if abs(dx/dy-1) > 0.001
    fprintf(1, 'x and y pixel sizes are different by %f, be warned\n', dx/dy-1);
end

% x range
indx = find( abs(x) <= diam/2 );
if mod(length(indx),2) ~= 0
    indx = indx(1:end-1);
end
indy = find( abs(y) <= diam/2 );
if mod(length(indy),2) ~= 0
    indy = indy(1:end-1);
end

dleng = length(indx) - length(indy);
if dleng > 0
    off = ceil( dleng/2 );
    indx = indx(off:off+length(indy)-1);
elseif dleng < 0
    off = ceil( -dleng/2 );
    indy = indy(off:off+length(indx)-1);
end

N = length(indx);

map = map(indy,indx);

if shape == 'c'
    w = hanning(N);
    m = w(:)*w(:).'; % Outer product window
    
    P_window =  (sum(sum(abs(m).^2))/N^2); % Power of the window
    
    map = map.*m/sqrt(P_window)/2;
end

ASD_2D = fftshift(fft2(ifftshift(map)));

rng2 = 2:2:N*2;

ASD_2D_L = zeros( N*2+2, N*2+2 );

ASD_2D_L( rng2+1, rng2 ) = ASD_2D/2;
ASD_2D_L( 1, rng2 ) = ASD_2D(end,:)/2;
ASD_2D_L( rng2-1, rng2 ) = ASD_2D_L( rng2-1, rng2 ) + ASD_2D/2;
ASD_2D_L( rng2, rng2+1 ) = ASD_2D/2;
ASD_2D_L( rng2, 1 ) = ASD_2D(:,end)/2;
ASD_2D_L( rng2, rng2-1 ) = ASD_2D_L( rng2, rng2-1 ) + ASD_2D/2;

ASD_2D_L( rng2+1, rng2+1 ) = ASD_2D_L( rng2+1, rng2   )/4;
ASD_2D_L( rng2+1, rng2+1 ) = ASD_2D_L( rng2+1, rng2+2 )/4 + ASD_2D_L( rng2+1, rng2+1 );
ASD_2D_L( rng2+1, rng2+1 ) = ASD_2D_L( rng2,   rng2+1 )/4 + ASD_2D_L( rng2+1, rng2+1 );
ASD_2D_L( rng2+1, rng2+1 ) = ASD_2D_L( rng2+2, rng2+1 )/4 + ASD_2D_L( rng2+1, rng2+1 );

ASD_2D_L( rng2, rng2 ) = ASD_2D;

ASD_2D_L = ifftshift( ASD_2D_L(2:end-1,2:end-1) );
ASD_2D_L(2:N,1) = (ASD_2D_L(2:N,1) + conj(ASD_2D_L(2*N:-1:N+2,1))) / 2;
ASD_2D_L(2*N:-1:N+2,1) = conj( ASD_2D_L(2:N,1) );
ASD_2D_L(1,1) = real( ASD_2D_L(1,1) );
ASD_2D_L(N+1,1) = real( ASD_2D_L(N+1,1) );

mapL = fftshift( ifft2( ASD_2D_L ) );

mapL = real( mapL );

xL = ((1:2*N)-N-0.5)*dx;
