function [EL, xlist] = propF( E0, W, L, ratio )
%
% stand alone simple FFT propagation code with magnification
% 
% EL = propF( E0, W, L, ratio )
%
% FFT propagation with adaptive grid sizing
%
% E0 is NxN field at the starting point
% W is the window size at the starting point
% L is the propagation distance
% ratio is the change of x-y scales, end scale / start scale
%
% EL is the field after propagating L
% xlist the x-coordinates at the end, i.e., EL(n,m) is at
% (xlist(n),xlist(m))
%

%
% June 7, 2017 copy of old propx, with support of even and odd N
%

N = size( E0, 1 );
k = 2*pi/1.064e-6;
u0 = (ratio -1) / L;
u1 = (ratio-1)*ratio / L;

% xlist = R/N * ( [ 0 : N-1 ] - (N-1)/2 );
% klist = 2*pi/R*[(0:N/2),(-N/2+1):-1];
xlist = W/N * ( [ 0 : N-1 ] - (N-1)/2 );
klist = 2*pi/W*[0:floor(N/2), floor(-N/2+1):-1];

% abs(E0(N/2, N/2-10:N/2+10))
E1 = E0 .*(  exp(1i*k/2*u0*xlist.^2).' * exp(1i*k/2*u0*xlist.^2) );

% abs(E1(N/2, N/2-10:N/2+10))
F1 = fft2( E1 );
% abs(F1(N/2, N/2-10:N/2+10))

FL = F1 .* ( exp(1i/(2*k)*(L/ratio)*klist.^2).' * exp(1i/(2*k)*(L/ratio)*klist.^2) );
% abs(FL(N/2, N/2-10:N/2+10))

EL = ifft2( FL );
% abs(EL(N/2, N/2-10:N/2+10))

EL = EL / ratio .* ( exp(-1i*k/2*u1*xlist.^2).' * exp(-1i*k/2*u1*xlist.^2));
% abs(EL(N/2, N/2-10:N/2+10))

xlist = xlist * ratio;
